@@ -11,18 +11,18 @@ from itertools import izip
 from scipy.sparse.linalg import eigsh
 
 from .decorators import inplace, linear, real, square, symmetric
-from .core import (Operator, BlockRowOperator, BroadcastingOperator,
-                   CompositionOperator, DenseOperator, DiagonalOperator,
-                   HomothetyOperator, ReductionOperator, Variable, X, _pool)
+from .core import (
+    Operator, BlockRowOperator, BroadcastingBase, CompositionOperator,
+    DenseOperator, DiagonalOperator, HomothetyOperator, ReductionOperator,
+    _pool)
 from .memory import empty
 from .utils import (
-    cast, complex_dtype, float_dtype, ifirst, izip_broadcast, pi, tointtuple)
+    cast, complex_dtype, float_dtype, izip_broadcast, pi, strshape, tointtuple)
 
 __all__ = [
     'BandOperator',
     'DegreesOperator',
     'DiagonalNumexprOperator',
-    'DiagonalNumexprNonSeparableOperator',
     'DifferenceOperator',
     'EigendecompositionOperator',
     'IntegrationTrapezeOperator',
@@ -73,20 +73,30 @@ class DiagonalNumexprOperator(DiagonalOperator):
     >>> d = DiagonalNumexprOperator(alpha, '(x/x0)**data', {'x':1.2, 'x0':1.})
 
     """
-    def __init__(self, data, expr, global_dict=None, var='data', dtype=float,
-                 **keywords):
+    def __init__(self, data, expr, global_dict=None, var='data',
+                 broadcast=None, dtype=None, **keywords):
         if not isinstance(expr, str):
             raise TypeError('The second argument is not a string expression.')
         if numexpr.__version__ < '2.0.2':
             keywords['flags'] = self.validate_flags(keywords.get('flags', {}),
                                                     inplace=False)
-        BroadcastingOperator.__init__(self, data, dtype=dtype, **keywords)
+        data = np.asarray(data)
+        if broadcast is None:
+            broadcast = 'scalar' if data.ndim == 0 else 'disabled'
+        if broadcast == 'disabled':
+            keywords['shapein'] = data.shape
+        if dtype is None:
+            dtype = float_dtype(data.dtype)
+
         self.expr = expr
         self.var = var
         self.global_dict = global_dict
         self._global_dict = {} if global_dict is None else global_dict.copy()
-        self._global_dict[var] = self.data.T if self.broadcast == \
-            'rightward' else self.data
+        self._global_dict[var] = data.T \
+            if broadcast is not None and broadcast.lower() == 'rightward' \
+            else data
+        BroadcastingBase.__init__(self, data, broadcast, dtype=dtype,
+                                  **keywords)
 
     def direct(self, input, output):
         if self.broadcast == 'rightward':
@@ -110,66 +120,6 @@ class DiagonalNumexprOperator(DiagonalOperator):
             self.expr, global_dict=self.global_dict, var=self.var)
 
 
-class DiagonalNumexprNonSeparableOperator(DiagonalOperator):
-    """
-    DiagonalOperator whose diagonal elements are calculated on the fly using
-    the numexpr package.
-
-    Notes
-    -----
-    - When such an instance is added or multiplied to another DiagonalOperator
-    (or subclass, such as an instance of this class), an algebraic
-    simplification takes place, which results in a regular (dense) diagonal
-    operator.
-    - This operator can not be separated so that each part handles a block
-    of a block operator. Also, rightward broadcasting cannot be used. If one of
-    these properties is desired, use the class DiagonalNumexprOperator.
-    - If the operator's input shape is not specified, its inference costs
-    an evaluation of the expression.
-
-    Example
-    -------
-    >>> alpha = np.arange(100.)
-    >>> d = DiagonalNumexprNonSeparableOperator(
-    ...     '(x/x0)**alpha', {'alpha': alpha, 'x': 1.2,'x0': 1})
-
-    """
-    def __init__(self, expr, global_dict=None, dtype=float, **keywords):
-        if not isinstance(expr, str):
-            raise TypeError('The first argument is not a string expression.')
-        if 'broadcast' in keywords and keywords['broadcast'] == 'rightward':
-            raise ValueError(
-                'The class DiagonalNumexprNonSeparableOperator does not handle'
-                ' rightward broadcasting. Use the class DiagonalNumexprOperato'
-                'r for this purpose.')
-        if 'broadcast' not in keywords or keywords['broadcast'] != 'leftward':
-            keywords['broadcast'] = 'disabled'
-        self.expr = expr
-        self.global_dict = global_dict
-        if 'shapein' not in keywords and 'shapeout' not in keywords and \
-           keywords['broadcast'] == 'disabled':
-            keywords['shapein'] = self.get_data().shape
-        if numexpr.__version__ < '2.0.2':
-            keywords['flags'] = self.validate_flags(keywords.get('flags', {}),
-                                                    inplace=False)
-        BroadcastingOperator.__init__(self, 0, dtype=dtype, **keywords)
-
-    def direct(self, input, output):
-        numexpr.evaluate('(' + self.expr + ') * input',
-                         global_dict=self.global_dict, out=output)
-
-    def get_data(self):
-        return numexpr.evaluate(self.expr, global_dict=self.global_dict)
-
-    @staticmethod
-    def _rule_left_block(self, op, cls):
-        return None
-
-    @staticmethod
-    def _rule_right_block(self, op, cls):
-        return None
-
-
 @real
 class IntegrationTrapezeOperator(BlockRowOperator):
     """
@@ -224,41 +174,118 @@ class IntegrationTrapezeOperator(BlockRowOperator):
 
 @real
 @linear
-@inplace
-class PackOperator(Operator):
-    """
-    Convert an ndarray into a vector, under the control of a mask.
-    """
+class PackBase(BroadcastingBase):
+    def __init__(self, data, broadcast, **keywords):
+        self.n = np.sum(data)
+        BroadcastingBase.__init__(self, data, broadcast, **keywords)
 
-    def __init__(self, mask, **keywords):
-        self.mask = ~np.array(mask, np.bool8, copy=False)
-        Operator.__init__(self, shapein=self.mask.shape,
-                          shapeout=np.sum(self.mask), **keywords)
-        #XXX .T does not share the same mask...
-        self.set_rule('T', lambda s: UnpackOperator(~s.mask, dtype=s.dtype))
+    def _reshape_packed(self, shape):
+        if self.broadcast == 'rightward':
+            return self.data.shape + shape[1:]
+        return shape[:-1] + self.data.shape
+
+    def _reshape_unpacked(self, shape):
+        if self.broadcast == 'rightward':
+            return (self.n,) + shape[self.data.ndim:]
+        return shape[:-self.data.ndim] + (self.n,)
+
+    @staticmethod
+    def _rule_broadcast(b1, b2, operation):
+        return
+
+    def _validate_packed(self, shape):
+        actual = shape[0 if self.broadcast == 'rightward' else -1]
+        if actual != self.n:
+            raise ValueError(
+                "The shape '{0}' is incompatible with that expected '{1}'.".
+                format(strshape(shape),
+                       strshape((self.n,), broadcast=self.broadcast)))
+
+    def _validate_unpacked(self, shape):
+        if self.broadcast == 'rightward':
+            actual = shape[:self.data.ndim]
+        else:
+            actual = shape[-self.data.ndim:]
+        if actual != self.data.shape:
+            raise ValueError(
+                "The shape '{0}' is incompatible with that expected '{1}'.".
+                format(strshape(shape),
+                       strshape(self.data.shape, broadcast=self.broadcast)))
+
+
+class PackOperator(PackBase):
+    """
+    Pack an ndarray into a vector under the control of a boolean mask.
+    The value True means that the element is kept.
+
+    """
+    def __init__(self, data, broadcast='disabled', **keywords):
+        data = np.array(data, bool, copy=False)
+        if np.all(data == data.flat[0]):
+            if data.flat[0]:
+                self.__class__ = DiagonalOperator
+                self.__init__(data, broadcast=broadcast, **keywords)
+                return
+        if broadcast.lower() == 'disabled':
+            keywords['shapein'] = data.shape
+            keywords['shapeout'] = np.sum(data)
+        PackBase.__init__(self, data, broadcast,
+                          reshapein=self._reshape_unpacked,
+                          reshapeout=self._reshape_packed,
+                          validatein=self._validate_unpacked,
+                          validateout=self._validate_packed, **keywords)
+        self.set_rule('T', lambda s: UnpackOperator(s.data,
+                                                    broadcast=s.broadcast))
         self.set_rule('T,.', '1', CompositionOperator)
 
     def direct(self, input, output):
-        output[...] = input[self.mask]
+        if self.broadcast == 'rightward':
+            output[...] = input[self.data, ...]
+        else:
+            output[...] = input[..., self.data]
+
+    @staticmethod
+    def _rule_left_block(op, self):
+        return
 
 
-@real
-@linear
-class UnpackOperator(Operator):
-    """
-    Convert a vector into an ndarray, under the control of a mask.
+class UnpackOperator(PackBase):
     """
+    Unpack a vector into an ndarray under the control of a mask.
+    The value True means that the element is kept.
 
-    def __init__(self, mask, **keywords):
-        self.mask = ~np.array(mask, np.bool8, copy=False)
-        Operator.__init__(self, shapein=np.sum(self.mask),
-                          shapeout=self.mask.shape, **keywords)
-        self.set_rule('T', lambda s: PackOperator(~s.mask, dtype=s.dtype))
+    """
+    def __init__(self, data, broadcast='disabled', **keywords):
+        data = np.array(data, bool, copy=False)
+        if np.all(data == data.flat[0]):
+            if data.flat[0]:
+                self.__class__ = DiagonalOperator
+                self.__init__(data, broadcast=broadcast, **keywords)
+                return
+        if broadcast.lower() == 'disabled':
+            keywords['shapein'] = np.sum(data)
+            keywords['shapeout'] = data.shape
+        PackBase.__init__(self, data, broadcast,
+                          reshapein=self._reshape_packed,
+                          reshapeout=self._reshape_unpacked,
+                          validatein=self._validate_packed,
+                          validateout=self._validate_unpacked, **keywords)
+        self.set_rule('T', lambda s: PackOperator(s.data,
+                                                  broadcast=s.broadcast))
         self.set_rule('T,.', '1', CompositionOperator)
 
     def direct(self, input, output):
         output[...] = 0
-        output[self.mask] = input
+        if self.broadcast == 'rightward':
+            output[self.data, ...] = input
+        else:
+            output[..., self.data] = input
+
+    @staticmethod
+    def _rule_right_block(self, op, cls):
+        if cls is CompositionOperator:
+            return
+        return BroadcastingBase._rule_right_block(self, op, cls)
 
 
 class RadiansOperator(HomothetyOperator):
