@@ -1,7 +1,7 @@
 import numpy as np
 from .core import (
     Operator, BlockOperator, BlockRowOperator, CompositionOperator)
-from .decorators import real
+from .flags import real
 from .utils import MPI
 
 # these function may be overridden
